#ifndef SC_FIFO_H
#define SC_FIFO_H

#include <sc_event.h>

namespace sc_core {

class sc_fifo_base
{
    public:
    sc_fifo_base();

    sc_event fifo_data_read_event;
    sc_event fifo_data_written_event;
};

template <class T>
class sc_fifo: public sc_fifo_base
{
    public:

    // Constructors

    explicit sc_fifo( int size = 16 ):
        capacity(size), offset(0), size_rd(0)
    {
        data = new T[size];
    }

    explicit sc_fifo( const char* name, int size = 16):
        capacity(size), offset(0), size_rd(0)
    {
        data = new T[size];
    }

    // Destructors

    ~sc_fifo();

    // Read access

    T read()
    {
        while(size_rd == offset)
        {
            wait();
        }

        T t = data[size_rd - 1];

        size_rd--;
        return t;
    }

    void read( T& t)
    {
        t = read();
    }

    bool nb_read( T& t)
    {
        if(size_rd == offset)
        {
            return false;
        }

        T t = data[size_rd - 1];
        size_rd--;
        return true;
    }

    operator T ()
    {
        return read();
    }

    // Write access

    void write( const T& t)
    {
        while(size_rd == capacity)
        {
            wait();
        }

        for(int i = size_rd; size_rd >= 0; i++)
        {
           data[i] = data[i-1];
        }

        offset++;
        size_rd++;

        data[0] = t;
    }
    bool nb_write( const T& t)
    {
        if(size_rd == capacity)
        {
            return false;
        }

        for(int i = size_rd; size_rd >= 0; i++)
        {
           data[i] = data[i-1];
        }

        offset++;
        size_rd++;

        data[0] = t;

        return true;
    }

    sc_fifo<T>& operator= ( const T& t)
    {
        write(t);
        return *this;
    }

    int num_available() const
    {
        return size_rd - offset;
    }

    int num_free() const
    {
        return capacity-offset;
    }

    // Events checking

    const sc_event& data_read_event() const
    {
        return fifo_data_read_event;
    }
    const sc_event& data_written_event() const
    {
        return fifo_data_written_event;
    }

private:

    // Update

    void update()
    {
        offset = 0;
    }

    // Wait

    void wait();

    // Disabled
    sc_fifo( const sc_fifo<T>& );
    sc_fifo& operator= ( const sc_fifo<T>& );

    T* data;
    const int capacity;
    int offset;
    int size_rd;

    bool data_written;
    bool data_read;
};

}

#endif
