#ifndef SC_FIFO_PORTS_H
#define SC_FIFO_PORTS_H

#include <sc_fifo.h>

namespace sc_core {

class sc_fifo_port_base
{
    public:

    sc_fifo_port_base():fifo(0)
    {
        scoot_fifo_port_base(0);
    }

    explicit sc_fifo_port_base(char * name):fifo(0)
    {
        scoot_fifo_port_base(name);
    }

    protected:

    // Port binding
    void sc_port_to_fifo_binding(sc_fifo_base& fifo);
    void sc_port_to_fifo_port_binding(sc_fifo_port_base& port);

    sc_fifo_base* fifo;

    private:
    sc_fifo_port_base(const sc_fifo_port_base&); // Disabled
    void scoot_fifo_port_base(char* name); // function marker for scoot
};

template <class T>
class sc_fifo_in: sc_fifo_port_base
{
    public:

    sc_fifo_in(){};
    explicit sc_fifo_in( const char* ){};

    void read( T& t)
    {
        ((sc_fifo<T>*)fifo)->read(t);
    }

    T read()
    {
        return ((sc_fifo<T>*)fifo)->read();
    }

    bool nb_read( T& t)
    {
        return ((sc_fifo<T>*)fifo)->nb_read(t);
    }

    const sc_event& data_written_event() const
    {
        return fifo_data_written_event;
    }

    int num_available() const
    {
        return ((sc_fifo<T>*)fifo)->num_available();
    }

    void operator()(sc_fifo<T>& f)
    { sc_port_to_fifo_binding(f); }

    void operator()(sc_fifo_in<T>& p)
    { sc_port_to_fifo_port_binding(p); }

    private:
    // Disabled
    sc_fifo_in( const sc_fifo_in<T>& );
    sc_fifo_in<T>& operator= ( const sc_fifo_in<T>& );

    sc_event fifo_data_written_event;
};

template <class T>
class sc_fifo_out: sc_fifo_port_base
{
    public:

    sc_fifo_out(){};
    explicit sc_fifo_out( const char* ){};

    void write(const T& t)
    {
        ((sc_fifo<T>*)fifo)->write(t);
    }

    bool nb_write(const T& t)
    {
        return ((sc_fifo<T>*)fifo)->nb_write(t);
    }

    const sc_event& data_read_event() const
    {
        return fifo_data_read_event;
    }

    int num_free() const
    {
        return ((sc_fifo<T>*)fifo)->num_free();
    }

    void operator()(sc_fifo<T>& f)
    { sc_port_to_fifo_binding(f); }

    void operator()(sc_fifo_out<T>& p)
    { sc_port_to_fifo_port_binding(p); }

    private:
    // Disabled
    sc_fifo_out( const sc_fifo_out<T>& );
    sc_fifo_out<T>& operator= ( const sc_fifo_out<T>& );

    sc_event fifo_data_read_event;
};

} // namespace sc_core

#endif
