#ifndef SC_SIGNAL_PORTS_H
#define SC_SIGNAL_PORTS_H

#include <sc_signal.h>

namespace sc_core {



class sc_port_base
{
    public:

    const sc_port_base& pos() const;
    const sc_port_base& neg() const;

    sc_port_base():signal(0)
    {
        scoot_port_base(0);
    }

    explicit sc_port_base(char * name):signal(0)
    {
        scoot_port_base(name);
    }

    protected:

    // Port binding
    void sc_port_to_signal_binding(sc_signal_base& signal);

    void sc_port_to_port_binding(sc_port_base&);

    sc_signal_base* signal;

    private:
    sc_port_base(const sc_port_base&); // Disabled

    void scoot_port_base(char* name); // function marker for scoot

};

template <class T>
class sc_port: public sc_port_base
{
    public:

    // Constructors

    sc_port(){};
    explicit sc_port(char* name):sc_port_base(name){};

    // Syntaxic sugar

    void operator()(sc_signal<T>& sig)
    { sc_port_to_signal_binding(sig); }

    void operator()(sc_port<T>& port)
    { sc_port_to_port_binding(port); }

    bool event() const
    { return ((const sc_signal<T>*) signal)->event(); }

    private:

    // Disabled

    sc_port(const sc_port&);

};


template <>
class sc_port<bool>: public sc_port_base
{
    public:

    // Constructors

    sc_port(){};
    explicit sc_port(char* name):sc_port_base(name){};

    // Syntaxic sugar

    void operator()(sc_signal<bool>& sig)
    { sc_port_to_signal_binding(sig); }

    void operator()(sc_port<bool>& port)
    { sc_port_to_port_binding(port); }

    bool event() const
    { return ((const sc_signal<bool>*) signal)->event(); }

    bool posedge() const
    { return ((const sc_signal<bool>*) signal)->posedge(); }

    bool negedge() const
    { return ((const sc_signal<bool>*) signal)->negedge(); }

    private:

    // Disabled

    sc_port(const sc_port&);

};


template <class T>
class sc_in: public sc_port<T>
{
    public:

    // Constructors

    sc_in(){};
    explicit sc_in(const char*name){};

    // Read access

    T read() const  {
        return ((sc_signal<T>*)signal)->read();
    }

    // Syntaxic sugar

    operator T () const
    { return read();}

    private:

    // Disabled

    sc_in(const sc_in&);

};

template <class T>
class sc_inout: public sc_port<T>
{
    public:

    // Constructors

    sc_inout(){};
    explicit sc_inout(const char* name){};

    // Access methods

    T read() const {
        return ((sc_signal<T>*)signal)->read();
    }

    void write(T t) {
        ((sc_signal<T>*)signal)->write(t);
    }

    // Syntaxic sugar
    operator T () const
    { return read(); }

    sc_inout<T>& operator = (T value)
    { write(value); return *this; }

    private:

    // Disabled

    sc_inout(const sc_inout&);
};

#define sc_out sc_inout



typedef sc_in<sc_logic> sc_in_resolved;
typedef sc_out<sc_logic> sc_out_resolved;
typedef sc_inout<sc_logic> sc_inout_resolved;

template <int W>
class sc_in_rv: public sc_in<sc_lv<W> >
{
    public:
    sc_in_rv(){};
    private:
    // Disabled
    sc_in_rv(const sc_in_rv&);
};


template <int W>
class sc_out_rv: public sc_out<sc_lv<W> >
{
    public:
    sc_out_rv(){};

    private:
    // Disabled
    sc_out_rv(const sc_out_rv&);
};

template <int W>
class sc_inout_rv: public sc_inout<sc_lv<W> >
{
    public:
    sc_inout_rv(){};

    private:
    // Disabled
    sc_inout_rv(const sc_inout_rv&);
};


} // namespace sc_core

#endif
